@@ -28,7 +28,6 @@
 #include <string.h>
 #include <errno.h>
 #include <unistd.h>
-#define _GNU_SOURCE
 #include <getopt.h>
 
 #include <libxml/parser.h>
@@ -37,6 +36,8 @@
 #include <gtk/gtk.h>
 #include <glib-unix.h>
 
+#include <gcr/gcr.h>
+
 #include <libsecret/secret.h>
 
 #include "openconnect.h"
@@ -740,41 +741,101 @@ static char *openconnect_get_cert_details(struct openconnect_info *vpninfo,
 }
 #endif
 
+static void cert_dialog_cancel_clicked (GtkButton *btn, GtkDialog *dlg)
+{
+	gtk_dialog_response(dlg, GTK_RESPONSE_CANCEL);
+}
+
+static void cert_dialog_connect_clicked (GtkButton *btn, GtkDialog *dlg)
+{
+	gtk_dialog_response(dlg, GTK_RESPONSE_OK);
+}
+
 static gboolean user_validate_cert(cert_data *data)
 {
 	auth_ui_data *ui_data = _ui_data; /* FIXME global */
+	
+	char *prevent_invalid_cert;
+	gboolean invalid_cert_allowed;
+	
+	GtkWidget *dlg;
 	char *title;
-	GtkWidget *dlg, *text, *scroll;
-	GtkTextBuffer *buffer;
+	
+	GtkWidget *vbox, *hbox, *expander_hbox;
+	
+	char *warning_label_text;
+	GtkWidget *warning_label;
+	
+	unsigned char *der_cert;
+	int der_cert_size;
+	GcrCertificate *cert;
+	GcrCertificateWidget *cert_widget;
+	
+	GtkWidget *cancel_button;
+	GtkWidget *security_expander;
+	GtkWidget *connect_button;
+	
 	int result;
-
+	
 	title = get_title(data->ui_data->vpn_name);
-	dlg = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_QUESTION,
-				     GTK_BUTTONS_OK_CANCEL,
-	                             _("Certificate from VPN server “%s” failed verification.\n"
-			             "Reason: %s\nDo you want to accept it?"),
-			             openconnect_get_hostname(data->ui_data->vpninfo),
-			             data->reason);
-	gtk_window_set_skip_taskbar_hint(GTK_WINDOW(dlg), FALSE);
-	gtk_window_set_skip_pager_hint(GTK_WINDOW(dlg), FALSE);
+	dlg = gtk_dialog_new();
+	gtk_window_set_modal(GTK_WINDOW(dlg), true);
 	gtk_window_set_title(GTK_WINDOW(dlg), title);
-	gtk_window_set_default_size(GTK_WINDOW(dlg), 550, 600);
-	gtk_window_set_resizable(GTK_WINDOW(dlg), TRUE);
-	gtk_dialog_set_default_response(GTK_DIALOG(dlg), GTK_RESPONSE_CANCEL);
-
+	gtk_window_set_default_size(GTK_WINDOW(dlg), 350, 200);
 	g_free(title);
 
-	scroll = gtk_scrolled_window_new(NULL, NULL);
-	gtk_box_pack_start(GTK_BOX (gtk_dialog_get_content_area(GTK_DIALOG (dlg))), scroll, TRUE, TRUE, 0);
-	gtk_widget_show(scroll);
-
-	text = gtk_text_view_new();
-	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(text));
-	gtk_text_buffer_set_text(buffer, data->cert_details, -1);
-	gtk_text_view_set_editable(GTK_TEXT_VIEW(text), 0);
-	gtk_text_view_set_cursor_visible(GTK_TEXT_VIEW(text), FALSE);
-	gtk_container_add(GTK_CONTAINER(scroll), text);
-	gtk_widget_show(text);
+	vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 8);
+	gtk_box_pack_start(GTK_BOX (gtk_dialog_get_content_area(GTK_DIALOG (dlg))), vbox, TRUE, TRUE, 0);
+	gtk_container_set_border_width(GTK_CONTAINER(vbox), 8);
+	gtk_widget_show(vbox);
+	
+	warning_label_text = g_strconcat(_("<b>The certificate may be invalid or untrusted!</b>\n"),
+									 _("<b>Reason: "), data->reason, ".</b>",
+									 NULL);
+	warning_label = gtk_label_new(NULL);
+	gtk_label_set_markup(GTK_LABEL(warning_label), warning_label_text);
+	gtk_box_pack_start(GTK_BOX(vbox), GTK_WIDGET(warning_label), FALSE, FALSE, 0);
+	gtk_widget_set_halign(warning_label, GTK_ALIGN_START);
+	gtk_label_set_line_wrap(GTK_LABEL(warning_label), true);
+	gtk_widget_show(warning_label);
+	g_free(warning_label_text);
+	
+	der_cert_size = openconnect_get_peer_cert_DER(data->ui_data->vpninfo, &der_cert);
+	cert = gcr_simple_certificate_new_static(der_cert, der_cert_size);
+	cert_widget = gcr_certificate_widget_new(cert);
+	gtk_box_pack_start(GTK_BOX(vbox), GTK_WIDGET(cert_widget), FALSE, FALSE, 0);
+	gtk_widget_show(GTK_WIDGET(cert_widget));
+	
+	hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 4);
+	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
+	gtk_widget_show(hbox);
+	
+	cancel_button = gtk_button_new_with_mnemonic(_("_Cancel"));
+	gtk_box_pack_start(GTK_BOX(hbox), cancel_button, FALSE, FALSE, 0);
+	g_signal_connect(cancel_button, "clicked", G_CALLBACK(cert_dialog_cancel_clicked), dlg);
+	gtk_widget_show(cancel_button);
+	
+	prevent_invalid_cert = g_hash_table_lookup(ui_data->options,
+							NM_OPENCONNECT_KEY_PREVENT_INVALID_CERT);									
+	invalid_cert_allowed = prevent_invalid_cert ? !strcmp(prevent_invalid_cert, "no") : TRUE;
+	
+	if (invalid_cert_allowed) {
+		security_expander = gtk_expander_new(_("I really know what I am doing"));
+		gtk_box_pack_start(GTK_BOX(vbox), security_expander, FALSE, FALSE, 0);
+		gtk_widget_show(security_expander);
+	
+		expander_hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
+		gtk_container_add(GTK_CONTAINER(security_expander), expander_hbox);
+		gtk_container_set_border_width(GTK_CONTAINER(expander_hbox), 0);
+		gtk_widget_show(expander_hbox);
+	
+		connect_button = gtk_button_new_with_label(_("Connect anyway"));
+		gtk_box_pack_start(GTK_BOX(expander_hbox), connect_button, FALSE, FALSE, 0);
+		gtk_widget_set_margin_start(connect_button, 12);
+		gtk_widget_set_margin_top(connect_button, 8);
+		g_signal_connect(connect_button, "clicked", G_CALLBACK(cert_dialog_connect_clicked), dlg);
+		gtk_widget_show(connect_button);
+	}
 
 	result = gtk_dialog_run(GTK_DIALOG(dlg));
 
@@ -894,7 +955,8 @@ static int parse_xmlconfig(gchar *xmlconfig)
 	}
 
 	xml_doc = xmlReadMemory(xmlconfig, strlen(xmlconfig), "noname.xml", NULL, 0);
-
+	if (!xml_doc)
+		return -EIO;
 	xml_node = xmlDocGetRootElement(xml_doc);
 	for (xml_node = xml_node->children; xml_node; xml_node = xml_node->next) {
                 if (xml_node->type == XML_ELEMENT_NODE &&
@@ -963,6 +1025,10 @@ static int get_config (auth_ui_data *ui_data,
 	char *csd;
 	char *sslkey, *cert;
 	char *csd_wrapper;
+	char *reported_os;
+#if OPENCONNECT_CHECK_VER(5,5)
+	char *key_pass;
+#endif
 	char *pem_passphrase_fsid;
 	char *cafile;
 	char *token_mode;
@@ -1011,12 +1077,12 @@ static int get_config (auth_ui_data *ui_data,
 	}
 
 	protocol = g_hash_table_lookup (options, NM_OPENCONNECT_KEY_PROTOCOL);
-	if (protocol && strcmp(protocol, "anyconnect")) {
 #if OPENCONNECT_CHECK_VER(5,1)
-		if (openconnect_set_protocol(vpninfo, protocol))
+	if (protocol && openconnect_set_protocol(vpninfo, protocol))
+#else
+	if (protocol && strcmp(protocol, "anyconnect"))
 #endif
-			return -EINVAL;
-	}
+		return -EINVAL;
 
 	cafile = g_hash_table_lookup (options, NM_OPENCONNECT_KEY_CACERT);
 	if (cafile)
@@ -1033,6 +1099,10 @@ static int get_config (auth_ui_data *ui_data,
 		openconnect_setup_csd(vpninfo, getuid(), 1, OC3DUP (csd_wrapper));
 	}
 
+	reported_os = g_hash_table_lookup (options, NM_OPENCONNECT_KEY_REPORTED_OS);
+	if (reported_os && reported_os[0])
+		openconnect_set_reported_os(vpninfo, reported_os);
+
 	proxy = g_hash_table_lookup (options, NM_OPENCONNECT_KEY_PROXY);
 	if (proxy && proxy[0] && openconnect_set_http_proxy(vpninfo, OC3DUP (proxy)))
 		return -EINVAL;
@@ -1041,6 +1111,12 @@ static int get_config (auth_ui_data *ui_data,
 	sslkey = g_hash_table_lookup (options, NM_OPENCONNECT_KEY_PRIVKEY);
 	openconnect_set_client_cert (vpninfo, OC3DUP (cert), OC3DUP (sslkey));
 
+#if OPENCONNECT_CHECK_VER(5,5)
+	key_pass = g_hash_table_lookup (options,
+					      NM_OPENCONNECT_KEY_KEY_PASS);
+	if (key_pass)
+		openconnect_set_key_password(vpninfo, key_pass);
+#endif
 	pem_passphrase_fsid = g_hash_table_lookup (options,
 						   NM_OPENCONNECT_KEY_PEM_PASSPHRASE_FSID);
 	if (pem_passphrase_fsid && cert && !strcmp(pem_passphrase_fsid, "yes"))
@@ -1559,7 +1635,8 @@ static auth_ui_data *init_ui_data (char *vpn_name, GHashTable *options, GHashTab
 							  g_free, g_free);
 	ui_data->success_passwords = g_hash_table_new_full (g_str_hash, g_str_equal,
 							  g_free, keyring_password_free);
-	ui_data->autosubmit = AUTOSUBMIT_LIMIT;
+	if (get_save_passwords (secrets))
+		ui_data->autosubmit = AUTOSUBMIT_LIMIT;
 
 	if (pipe(ui_data->cancel_pipes)) {
 		/* This should never happen, and the world is probably about
