@@ -53,8 +53,16 @@
 # include "nm-utils/nm-vpn-plugin-utils.h"
 #endif
 
-#define OPENCONNECT_PLUGIN_NAME    _("Cisco AnyConnect Compatible VPN (openconnect)")
-#define OPENCONNECT_PLUGIN_DESC    _("Compatible with Cisco AnyConnect SSL VPN.")
+#if OPENCONNECT_CHECK_VER(5,5)
+# define OPENCONNECT_PLUGIN_NAME    _("Multi-protocol VPN client (openconnect)")
+# define OPENCONNECT_PLUGIN_DESC    _("Compatible with Cisco AnyConnect, Juniper Network Connect and Junos Pulse, and PAN GlobalProtect SSL VPNs.")
+#elif OPENCONNECT_CHECK_VER(5,2)
+# define OPENCONNECT_PLUGIN_NAME    _("Multi-protocol VPN client (openconnect)")
+# define OPENCONNECT_PLUGIN_DESC    _("Compatible with Cisco AnyConnect and Juniper Network Connect and Junos Pulse SSL VPNs.")
+#else
+# define OPENCONNECT_PLUGIN_NAME    _("Cisco AnyConnect Compatible VPN (openconnect)")
+# define OPENCONNECT_PLUGIN_DESC    _("Compatible with Cisco AnyConnect SSL VPN.")
+#endif
 
 /************** plugin class **************/
 
@@ -71,8 +79,60 @@ G_DEFINE_TYPE_EXTENDED (OpenconnectEditorPlugin, openconnect_editor_plugin, G_TY
                         G_IMPLEMENT_INTERFACE (NM_TYPE_VPN_EDITOR_PLUGIN,
                                                openconnect_editor_plugin_interface_init))
 
+#if !OPENCONNECT_CHECK_VER(5,5)
+#define OC_PROTO_PROXY  (1<<0)
+#define OC_PROTO_CSD    (1<<1)
+#define OC_PROTO_AUTH_CERT      (1<<2)
+#define OC_PROTO_AUTH_OTP       (1<<3)
+#define OC_PROTO_AUTH_STOKEN    (1<<4)
+
+struct oc_vpn_proto {
+        const char *name;
+        const char *pretty_name;
+        const char *description;
+        unsigned int flags;
+};
+
+static int openconnect_get_supported_protocols(struct oc_vpn_proto **protos)
+{
+	struct oc_vpn_proto *pr;
+
+	*protos = pr = calloc(sizeof(*pr), 2);
+	if (!pr)
+		return -ENOMEM;
+
+	pr[0].name = "anyconnect";
+	pr[0].pretty_name = _("Cisco AnyConnect or openconnect");
+	pr[0].description = _("Compatible with Cisco AnyConnect SSL VPN, as well as ocserv");
+	pr[0].flags = OC_PROTO_PROXY | OC_PROTO_CSD | OC_PROTO_AUTH_CERT | OC_PROTO_AUTH_OTP | OC_PROTO_AUTH_STOKEN;
+
+	pr[1].name = "nc";
+	pr[1].pretty_name = _("Juniper Network Connect");
+	pr[1].description = _("Compatible with Juniper Network Connect");
+	pr[1].flags = OC_PROTO_PROXY | OC_PROTO_CSD | OC_PROTO_AUTH_CERT | OC_PROTO_AUTH_OTP;
+
+	/* Newer protocols like GlobalProtect and Pulse only came after
+	 * the openconnect_get_supported_protocols() API, so we don't need
+	 * hard-coded knowledge about those. */
+
+#if OPENCONNECT_CHECK_VER(5,2)
+	/* OpenConnect v7.05 (API 5.2) onwards had nc support. */
+	return 2;
+#else
+	/* Before that, only AnyConnect. */
+	return 1;
+#endif
+}
+
+static void openconnect_free_supported_protocols(struct oc_vpn_proto *protos)
+{
+	free(protos);
+}
+#endif /* !OPENCONNECT_CHECK_VER(5,5) */
+
 typedef struct {
-	char **supported_protocols;
+	int nr_supported_protocols;
+	struct oc_vpn_proto *supported_protocols;
 } OpenconnectEditorPluginPrivate;
 
 #define OPENCONNECT_EDITOR_PLUGIN_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), OPENCONNECT_TYPE_EDITOR_PLUGIN, OpenconnectEditorPluginPrivate))
@@ -153,9 +213,10 @@ import (NMVpnEditorPlugin *iface, const char *path, GError **error)
 	if (buf)
 		g_object_set (s_con, NM_SETTING_CONNECTION_ID, buf, NULL);
 
-	/* CA Certificate */
+	/* CA Certificate. We have an exception for filename "(null)" because
+	 * the exporter used to do that wrongly. */
 	buf = g_key_file_get_string (keyfile, "openconnect", "CACert", NULL);
-	if (buf)
+	if (buf && strcmp(buf, "(null)"))
 		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_CACERT, buf);
 
 	/* Protocol */
@@ -178,20 +239,30 @@ import (NMVpnEditorPlugin *iface, const char *path, GError **error)
 	if (buf)
 		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_CSD_WRAPPER, buf);
 
+	/* Reported OS */
+	buf = g_key_file_get_string (keyfile, "openconnect", "ReportedOS", NULL);
+	if (buf)
+		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_REPORTED_OS, buf);
+
 	/* User Certificate */
 	buf = g_key_file_get_string (keyfile, "openconnect", "UserCertificate", NULL);
-	if (buf)
+	if (buf && strcmp(buf, "(null)"))
 		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_USERCERT, buf);
 
 	/* Private Key */
 	buf = g_key_file_get_string (keyfile, "openconnect", "PrivateKey", NULL);
-	if (buf)
+	if (buf && strcmp(buf, "(null)"))
 		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_PRIVKEY, buf);
 
 	/* FSID */
 	bval = g_key_file_get_boolean (keyfile, "openconnect", "FSID", NULL);
 	if (bval)
 		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_PEM_PASSPHRASE_FSID, "yes");
+	
+	/* Prevent invalid cert */
+	bval = g_key_file_get_boolean (keyfile, "openconnect", "PreventInvalidCert", NULL);
+	if (true)
+		nm_setting_vpn_add_data_item (s_vpn, NM_OPENCONNECT_KEY_PREVENT_INVALID_CERT, "yes");
 
 	/* Soft token mode */
 	buf = g_key_file_get_string (keyfile, "openconnect", "StokenSource", NULL);
@@ -221,9 +292,11 @@ export (NMVpnEditorPlugin *iface,
 	const char *proxy = NULL;
 	gboolean csd_enable = FALSE;
 	const char *csd_wrapper = NULL;
+	const char *reported_os = NULL;
 	const char *usercert = NULL;
 	const char *privkey = NULL;
 	gboolean pem_passphrase_fsid = FALSE;
+	gboolean prevent_invalid_cert = FALSE;
 	const char *token_mode = NULL;
 	const char *token_secret = NULL;
 	gboolean success = FALSE;
@@ -273,6 +346,10 @@ export (NMVpnEditorPlugin *iface,
 	if (value && strlen (value))
 		csd_wrapper = value;
 
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_REPORTED_OS);
+	if (value && strlen (value))
+		reported_os = value;
+
 	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_USERCERT);
 	if (value && strlen (value))
 		usercert = value;
@@ -284,6 +361,10 @@ export (NMVpnEditorPlugin *iface,
 	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_PEM_PASSPHRASE_FSID);
 	if (value && !strcmp (value, "yes"))
 		pem_passphrase_fsid = TRUE;
+	
+	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_PREVENT_INVALID_CERT);
+	if (value && !strcmp (value, "yes"))
+		prevent_invalid_cert = TRUE;
 
 	value = nm_setting_vpn_get_data_item (s_vpn, NM_OPENCONNECT_KEY_TOKEN_MODE);
 	if (value && strlen (value))
@@ -307,21 +388,25 @@ export (NMVpnEditorPlugin *iface,
 		 "Proxy=%s\n"
 		 "CSDEnable=%s\n"
 		 "CSDWrapper=%s\n"
+		 "ReportedOS=%s\n"
 		 "UserCertificate=%s\n"
 		 "PrivateKey=%s\n"
 		 "FSID=%s\n"
+		 "PreventInvalidCert=%s\n"
 		 "StokenSource=%s\n"
 		 "StokenString=%s\n",
 		 /* Description */           nm_setting_connection_get_id (s_con),
 		 /* Host */                  gateway,
-		 /* CA Certificate */        cacert,
+		 /* CA Certificate */        cacert ? cacert : "",
 		 /* Protocol */              protocol ? protocol : "anyconnect",
 		 /* Proxy */                 proxy ? proxy : "",
 		 /* Cisco Secure Desktop */  csd_enable ? "1" : "0",
 		 /* CSD Wrapper Script */    csd_wrapper ? csd_wrapper : "",
-		 /* User Certificate */      usercert,
-		 /* Private Key */           privkey,
+		 /* Reported OS */           reported_os ? reported_os : "",
+		 /* User Certificate */      usercert ? usercert : "",
+		 /* Private Key */           privkey ? privkey : "",
 		 /* FSID */                  pem_passphrase_fsid ? "1" : "0",
+		 /* Prevent invalid cert */  prevent_invalid_cert ? "1" : "0",
 		 /* Soft token mode */       token_mode ? token_mode : "",
 		 /* Soft token secret */     token_secret ? token_secret : "");
 
@@ -347,35 +432,30 @@ notify_plugin_info_set (NMVpnEditorPlugin *plugin,
                         NMVpnPluginInfo *plugin_info)
 {
 	OpenconnectEditorPluginPrivate *priv = OPENCONNECT_EDITOR_PLUGIN_GET_PRIVATE (plugin);
-	const char *supported_protocols;
-	guint i, j;
 
 	if (!plugin_info)
 		return;
 
-	supported_protocols = nm_vpn_plugin_info_lookup_property (plugin_info, "openconnect", "supported-protocols");
+	openconnect_free_supported_protocols(priv->supported_protocols);
+	priv->supported_protocols = NULL;
 
-	g_strfreev (priv->supported_protocols);
-	priv->supported_protocols = supported_protocols
-	    ? g_strsplit_set (supported_protocols, ",", -1)
-	    : g_new0 (char *, 1);
-
-	/*remove empty entries and whitespace */
-	for (i = 0, j = 0; priv->supported_protocols[j]; j++) {
-		g_strstrip (priv->supported_protocols[j]);
-		if (priv->supported_protocols[j][0] == '\0')
-			g_free (priv->supported_protocols[j]);
-		else
-			priv->supported_protocols[i++] = priv->supported_protocols[j];
-	}
-	priv->supported_protocols[i] = NULL;
+	priv->nr_supported_protocols = openconnect_get_supported_protocols(&priv->supported_protocols);
+	if (priv->nr_supported_protocols > 0)
+		return;
 }
 
 static char **
 _vt_impl_get_service_add_details (NMVpnEditorPlugin *plugin,
                                   const char *service_type)
 {
-	return g_strdupv (OPENCONNECT_EDITOR_PLUGIN_GET_PRIVATE (plugin)->supported_protocols);
+	OpenconnectEditorPluginPrivate *priv = OPENCONNECT_EDITOR_PLUGIN_GET_PRIVATE (plugin);
+	guint i;
+	char **ret = calloc(sizeof(char *), priv->nr_supported_protocols + 1);
+
+	for (i = 0; i < priv->nr_supported_protocols; i++)
+		ret[i] = strdup(priv->supported_protocols[i].name);
+
+	return ret;
 }
 
 static gboolean
@@ -395,26 +475,18 @@ _vt_impl_get_service_add_detail (NMVpnEditorPlugin *plugin,
 		return FALSE;
 
 	priv = OPENCONNECT_EDITOR_PLUGIN_GET_PRIVATE (plugin);
-	for (i = 0; priv->supported_protocols[i]; i++) {
-		if (!nm_streq (add_detail, priv->supported_protocols[i]))
+	for (i = 0; i < priv->nr_supported_protocols; i++) {
+		struct oc_vpn_proto *p = &priv->supported_protocols[i];
+
+		if (!nm_streq (add_detail, p->name))
 			continue;
-		if (nm_streq (add_detail, "anyconnect")) {
-			NM_SET_OUT (out_pretty_name, g_strdup (OPENCONNECT_PLUGIN_NAME));
-			NM_SET_OUT (out_description, g_strdup (OPENCONNECT_PLUGIN_DESC));
-			/* we unset @add_detail, because "anyconnect" is the default and no need
-			 * to set the protocol explicitly. */
-			add_detail = NULL;
-		} else if (nm_streq (add_detail, "nc")) {
-			NM_SET_OUT (out_pretty_name, g_strdup (_("Juniper Network Connect (openconnect)")));
-			NM_SET_OUT (out_description, g_strdup (_("Compatible with Juniper Network Connect / Pulse Secure SSL VPN")));
-		} else {
-			/* we don't know this protocol by name, but it's configured in the .name file,
-			 * so just take it. */
-			NM_SET_OUT (out_pretty_name, g_strdup_printf (_("Openconnect VPN (%s)"), add_detail));
-			NM_SET_OUT (out_description, g_strdup_printf (_("Openconnect SSL VPN with %s protocol"), add_detail));
+		NM_SET_OUT (out_pretty_name, g_strdup_printf("%s (OpenConnect)", p->pretty_name));
+		NM_SET_OUT (out_description, g_strdup (p->description));
+
+		if (i) {
+			NM_SET_OUT (out_add_detail_key, g_strdup (add_detail ? NM_OPENCONNECT_KEY_PROTOCOL : NULL));
+			NM_SET_OUT (out_add_detail_val, g_strdup (add_detail));
 		}
-		NM_SET_OUT (out_add_detail_key, g_strdup (add_detail ? NM_OPENCONNECT_KEY_PROTOCOL : NULL));
-		NM_SET_OUT (out_add_detail_val, g_strdup (add_detail));
 		NM_SET_OUT (out_flags, 0);
 		return TRUE;
 	}
@@ -488,7 +560,7 @@ openconnect_editor_plugin_dispose (GObject *object)
 	OpenconnectEditorPlugin *plugin = OPENCONNECT_EDITOR_PLUGIN (object);
 	OpenconnectEditorPluginPrivate *priv = OPENCONNECT_EDITOR_PLUGIN_GET_PRIVATE (plugin);
 
-	g_strfreev (priv->supported_protocols);
+	openconnect_free_supported_protocols(priv->supported_protocols);
 	priv->supported_protocols = NULL;
 
 	G_OBJECT_CLASS (openconnect_editor_plugin_parent_class)->dispose (object);
@@ -520,14 +592,6 @@ openconnect_editor_plugin_class_init (OpenconnectEditorPluginClass *req_class)
 static void
 openconnect_editor_plugin_init (OpenconnectEditorPlugin *plugin)
 {
-	OpenconnectEditorPluginPrivate *priv = OPENCONNECT_EDITOR_PLUGIN_GET_PRIVATE (plugin);
-	char *dflt[] = {
-		"anyconnect",
-		"nc",
-		NULL,
-	};
-
-	priv->supported_protocols = g_strdupv (dflt);
 }
 
 static void
